# -*- encoding: utf-8 -*-
from pyspark.ml.regression import LinearRegression, RandomForestRegressor
from sklearn.metrics import r2_score, mean_absolute_error, mean_squared_error

from .base import BasePipelineBuilder
from sparkmlpipe.utils.constants import LINEAR_REGRESSION_REGRESSOR, RANDOM_FOREST_REGRESSOR
from sparkmlpipe.utils.exceptions import InvalidConfigError


class RegressionPipelineBuilder(BasePipelineBuilder):
    def get_metrics(self, y_true, y_pred):
        metrics = dict()

        metrics['r2_score'] = r2_score(y_true, y_pred)
        metrics['mean_squared_error'] = mean_squared_error(y_true, y_pred)
        metrics['mean_absolute_error'] = mean_absolute_error(y_true, y_pred)

        return metrics

    def _get_model_stages(self):
        model_stages = []
        model_conf = self.config['pipe']['model']
        params = model_conf['params'] if 'params' in model_conf else {}
        if model_conf['type'] == LINEAR_REGRESSION_REGRESSOR:
            model = LinearRegression(featuresCol=self.feat_col, labelCol=self.label_col)
            if 'maxIter' in params:
                model.setMaxIter(params['maxIter'])
            if 'regParam' in params:
                model.setRegParam(params['regParam'])
            if 'elasticNetParam' in params:
                model.setElasticNetParam(params['elasticNetParam'])

        elif model_conf['type'] == RANDOM_FOREST_REGRESSOR:
            model = RandomForestRegressor(featuresCol=self.feat_col, labelCol=self.label_col)
            if 'seed' in params:
                model.setSeed(params['seed'])
            if 'numTrees' in params:
                model.setNumTrees(params['numTrees'])
            if 'maxDepth' in params:
                model.setMaxDepth(params['maxDepth'])
            if 'subsamplingRate' in params:
                model.setSubsamplingRate(params['subsamplingRate'])
        else:
            raise InvalidConfigError('pipe.model.type', model_conf['type'])

        model_stages.append(model)
        return model_stages



# -*- encoding: utf-8 -*-
import os
from unittest import TestCase

import pyspark
from pyspark.sql import SparkSession


from sparkmlpipe.sparkmlpipe import SparkStatPipeline
from sparkmlpipe.utils.common import load_config
from sparkmlpipe.utils.constants import ROOT_PATH, APP_PATH, CONFIG_CORR_FILE, CONFIG_CORR_HANGYAN_FILE


class TestStat(TestCase):

    spark = SparkSession.builder \
        .master("local") \
        .appName("Test") \
        .getOrCreate()

    def test_corr_iris(self):
        conf_path = os.path.join(APP_PATH, CONFIG_CORR_FILE)
        config = load_config(conf_path)

        data_path = config['data']['location']

        data = self.spark.read.option('header', 'true') \
            .option("inferSchema", "true") \
            .csv(os.path.join(ROOT_PATH, data_path))

        spark_stat_pipe = SparkStatPipeline(config, data)

        corr = spark_stat_pipe.get_stat()

        self.assertTrue(isinstance(corr, pyspark.ml.linalg.DenseMatrix))

    def test_corr_hangyan(self):
        conf_path = os.path.join(APP_PATH, CONFIG_CORR_HANGYAN_FILE)
        config = load_config(conf_path)

        data_path = config['data']['location']

        data = self.spark.read.option('header', 'true') \
            .option("inferSchema", "true") \
            .csv(os.path.join(ROOT_PATH, data_path))

        spark_stat_pipe = SparkStatPipeline(config, data)

        corr = spark_stat_pipe.get_stat()

        self.assertTrue(isinstance(corr, pyspark.ml.linalg.DenseMatrix))


# -*- encoding: utf-8 -*-
from os.path import dirname

LOGGING_CONF = 'conf/logging.yaml'
CONFIG_REGRESSOR_LINEAR_FILE = 'conf/conf_regressor_linear.yaml'
CONFIG_CLASSIFY_LOG_FILE = 'conf/conf_classify_log.yaml'
CONFIG_REGRESSOR_RF_FILE = 'conf/conf_regressor_rf.yaml'
CONFIG_CLASSIFY_RF_FILE = 'conf/conf_classify_rf.yaml'
CONFIG_CORR_FILE = 'conf/conf_corr.yaml'
CONFIG_CORR_HANGYAN_FILE = 'conf/conf_corr_hangyan.yaml'


ROOT_PATH = dirname(dirname(dirname(__file__)))
APP_PATH = dirname(dirname(__file__))

STRING_TYPE = 'string'
NUM_TYPES = ['vector', 'integer', 'long', 'double', 'float']

# scale types
MIN_MAX_SCALER_TYPE = 1
STANDARD_SCALER_TYPE = 2
MAXABS_SCALER_TYPE = 3

# impute types
MEAN_IMPUTER_TYPE = 1
MEDIAN_IMPUTER_TYPE = 2

# feature prep types
CHISQ_FEAT_TYPE = 1
PCA_FEAT_TYPE = 2
DEFAULT_K = 2

# pipe types
CLASSIFICATION = 1
REGRESSION = 2
CORRELATION = 3

# classifier types
LOGISTIC_REGRESSION_CLASSIFIER = 1
RANDOM_FOREST_CLASSIFIER = 2

# regressor types
LINEAR_REGRESSION_REGRESSOR = 1
RANDOM_FOREST_REGRESSOR = 2

# correlation types
PEARSON_CORR = 1
SPEARMAN_CORR = 2



# -*- encoding: utf-8 -*-
import pyspark

from .constants import CLASSIFICATION, REGRESSION, CORRELATION
from .constants import RANDOM_FOREST_CLASSIFIER, RANDOM_FOREST_REGRESSOR
from .exceptions import MismatchColumnNumberError, InvalidConfigError, MissingConfigError, MismatchTypeError

from sparkmlpipe.pipeline.classification import ClassificationPipelineBuilder
from sparkmlpipe.pipeline.regression import RegressionPipelineBuilder
from sparkmlpipe.stat.correlation import CorrelationBuilder


# general config file check
def verify_config(config):
    assert isinstance(config, dict)

    if 'data' not in config.keys():
        raise MissingConfigError('data')
    else:
        for key in ['location', 'header', 'types']:
            if key not in config['data'].keys():
                raise MissingConfigError('data.' + key)
        # check if header and types has the same size
        num_header_cols = len(config['data']['header'])
        num_types_cols = len(config['data']['types'])
        if num_header_cols != num_types_cols:
            raise MismatchColumnNumberError(num_header_cols, num_types_cols)

    if 'pipe' not in config.keys():
        raise MissingConfigError('pipe')
    else:
        for key in ['type', 'data_prep', 'feat_prep', 'model']:
            if key not in config['pipe'].keys():
                raise MissingConfigError('pipe.' + key)


def verify_input_dataset(config, data):
    num_conf_cols = len(config['data']['types'])
    num_data_cols = len(data.columns)

    if num_conf_cols > num_data_cols:
        raise MismatchColumnNumberError(num_conf_cols, num_data_cols)


def get_pipeline_builder(pipe_type):
    if pipe_type == CLASSIFICATION:
        return ClassificationPipelineBuilder
    elif pipe_type == REGRESSION:
        return RegressionPipelineBuilder
    elif pipe_type == CORRELATION:
        return CorrelationBuilder
    else:
        raise InvalidConfigError('pipe.type', pipe_type)


def fit_pipeline(pipeline, config, data, label_col):
    if isinstance(pipeline, pyspark.ml.Pipeline):
        pipe_type = config['pipe']['type']
        model_type = config['pipe']['model']['type']

        model_pipe = pipeline.fit(data)
        predictions = model_pipe.transform(data)

        predictions_df = predictions.select(label_col, 'prediction').toPandas()
        y_true = predictions_df[label_col]
        y_pred = predictions_df['prediction']

        feature_importance = None

        # NOTE: toPandas() should only be used if the resulting Pandas's DataFrame is expected to be small,
        # as all the data is loaded into the driver's memory.
        if pipe_type == CLASSIFICATION:
            if model_type == RANDOM_FOREST_CLASSIFIER:
                feature_importance = model_pipe.stages[-1].featureImportances
        elif pipe_type == REGRESSION:
            if model_type == RANDOM_FOREST_REGRESSOR:
                feature_importance = model_pipe.stages[-1].featureImportances
        else:
            raise InvalidConfigError('pipe.type', pipe_type)
    else:
        raise MismatchTypeError(type(pipeline), pyspark.ml.Pipeline)

    return model_pipe, y_true, y_pred, feature_importance



# -*- encoding: utf-8 -*-
import logging
import logging.config
import os
import yaml

from .constants import LOGGING_CONF


def setup_logger(output_file=None):
    with open(os.path.join(os.path.dirname(os.path.dirname(__file__)), LOGGING_CONF), 'r') as fh:
        config = yaml.load(fh)
    if output_file is not None:
        config['handlers']['file_handler']['filename'] = output_file
    logging.config.dictConfig(config)


def _create_logger(name):
    logging.basicConfig(format='[%(levelname)s] [%(asctime)s:%(name)s] %('
                           'message)s', datefmt='%H:%M:%S')
    return logging.getLogger(name)


def get_logger(name):
    logger = PickableLoggerAdapter(name)
    return logger


class PickableLoggerAdapter(object):

    def __init__(self, name):
        self.name = name
        self.logger = _create_logger(name)

    def __getstate__(self):
        """
        Method is called when pickle dumps an object.

        Returns
        -------
        Dictionary, representing the object state to be pickled. Ignores
        the self.logger field and only returns the logger name.
        """
        return { 'name': self.name }

    def __setstate__(self, state):
        """
        Method is called when pickle loads an object. Retrieves the name and
        creates a logger.

        Parameters
        ----------
        state - dictionary, containing the logger name.

        """
        self.name = state['name']
        self.logger = _create_logger(self.name)

    def debug(self, msg, *args, **kwargs):
        self.logger.debug(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self.logger.info(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.logger.warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self.logger.error(msg, *args, **kwargs)

    def exception(self, msg, *args, **kwargs):
        self.logger.exception(msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        self.logger.critical(msg, *args, **kwargs)

    def log(self, level, msg, *args, **kwargs):
        self.logger.log(level, msg, *args, **kwargs)

    def isEnabledFor(self, level):
        return self.logger.isEnabledFor(level)



# -*- encoding: utf-8 -*-
import pkg_resources
import re
from distutils.version import LooseVersion

from .exceptions import MissingPackageError, IncorrectPackageVersionError

SUBPATTERN = r'((?P<operation%d>==|>=|>|<)(?P<version%d>(\d+)?(\.[a-zA-Z0-9]+)?(\.\d+)?))'
RE_PATTERN = re.compile(
    r'^(?P<name>[\w\-]+)%s?(,%s)?$' % (SUBPATTERN % (1, 1), SUBPATTERN % (2, 2)))


def verify_packages(packages):
    if not packages:
        return
    if isinstance(packages, str):
        packages = packages.splitlines()

    for package in packages:
        if not package:
            continue

        match = RE_PATTERN.match(package)
        if match:
            name = match.group('name')
            operation = match.group('operation1')
            version = match.group('version1')
            _verify_package(name, operation, version)
            operation2 = match.group('operation2')
            version2 = match.group('version2')
        else:
            raise ValueError('Unable to read requirement: %s' % package)


def _verify_package(name, operation, version):
    try:
        module = pkg_resources.get_distribution(name)
    except pkg_resources.DistributionNotFound:
        raise MissingPackageError(name)

    if not operation:
        return

    required_version = LooseVersion(version)
    installed_version = LooseVersion(module.version)

    if operation == '==':
        check = required_version == installed_version
    elif operation == '>':
        check = installed_version > required_version
    elif operation == '>=':
        check = installed_version > required_version or \
                installed_version == required_version
    else:
        raise NotImplementedError(
            'operation \'%s\' is not supported' % operation)
    if not check:
        raise IncorrectPackageVersionError(name, installed_version, operation,
                                           required_version)






# -*- encoding: utf-8 -*-
import os
from unittest import TestCase

import pyspark
from pyspark.sql import SparkSession

from sparkmlpipe.sparkmlpipe import SparkMLPipeline
from sparkmlpipe.utils.common import load_config
from sparkmlpipe.utils.constants import ROOT_PATH, APP_PATH, CONFIG_CLASSIFY_LOG_FILE, CONFIG_CLASSIFY_RF_FILE,\
    CONFIG_REGRESSOR_LINEAR_FILE, CONFIG_REGRESSOR_RF_FILE


class TestPipe(TestCase):

    spark = SparkSession.builder \
        .master("local") \
        .appName("Test") \
        .getOrCreate()

    def test_classify_log_pipe(self):
        conf_path = os.path.join(APP_PATH, CONFIG_CLASSIFY_LOG_FILE)
        config = load_config(conf_path)

        data_path = config['data']['location']

        data = self.spark.read.option('header', 'true')\
            .option("inferSchema", "true") \
            .csv(os.path.join(ROOT_PATH, data_path))

        sparkml_pipe = SparkMLPipeline(config)

        model, y_true, y_pred, _ = sparkml_pipe.fit_with_outputs(data)

        metrics = sparkml_pipe.get_metrics(y_true, y_pred)

        self.assertTrue(len(model.stages) == 9)
        self.assertTrue(len(y_true) == len(y_pred))
        self.assertTrue(len(metrics) == 4)

    def test_classify_rf_pipe(self):
        conf_path = os.path.join(APP_PATH, CONFIG_CLASSIFY_RF_FILE)
        config = load_config(conf_path)

        data_path = config['data']['location']

        data = self.spark.read.option('header', 'true')\
            .option("inferSchema", "true") \
            .csv(os.path.join(ROOT_PATH, data_path))

        sparkml_pipe = SparkMLPipeline(config)

        model, y_true, y_pred, feature_importances = sparkml_pipe.fit_with_outputs(data)

        metrics = sparkml_pipe.get_metrics(y_true, y_pred)

        self.assertTrue(len(model.stages) == 10)
        self.assertTrue(len(y_true) == len(y_pred))
        self.assertTrue(len(metrics) == 4)
        self.assertTrue(isinstance(feature_importances, pyspark.ml.linalg.SparseVector))

    def test_regressor_linear_pipe(self):
        conf_path = os.path.join(APP_PATH, CONFIG_REGRESSOR_LINEAR_FILE)
        config = load_config(conf_path)

        data_path = config['data']['location']

        data = self.spark.read.option('header', 'true')\
            .option("inferSchema", "true") \
            .csv(os.path.join(ROOT_PATH, data_path))

        sparkml_pipe = SparkMLPipeline(config)

        model, y_true, y_pred, _ = sparkml_pipe.fit_with_outputs(data)

        metrics = sparkml_pipe.get_metrics(y_true, y_pred)

        self.assertTrue(len(model.stages) == 9)
        self.assertTrue(len(y_true) == len(y_pred))
        self.assertTrue(len(metrics) == 3)

    def test_regressor_rf_pipe(self):
        conf_path = os.path.join(APP_PATH, CONFIG_REGRESSOR_RF_FILE)
        config = load_config(conf_path)

        data_path = config['data']['location']

        data = self.spark.read.option('header', 'true')\
            .option("inferSchema", "true") \
            .csv(os.path.join(ROOT_PATH, data_path))

        sparkml_pipe = SparkMLPipeline(config)

        model, y_true, y_pred, feature_importances = sparkml_pipe.fit_with_outputs(data)

        metrics = sparkml_pipe.get_metrics(y_true, y_pred)

        self.assertTrue(len(model.stages) == 10)
        self.assertTrue(len(y_true) == len(y_pred))
        self.assertTrue(len(metrics) == 3)
        self.assertTrue(isinstance(feature_importances, pyspark.ml.linalg.SparseVector))



# -*- encoding: utf-8 -*-
from pyspark.ml import Pipeline

from sparkmlpipe.utils.pipeline import fit_pipeline, get_pipeline_builder, verify_config, verify_input_dataset


class SparkMLPipeline(Pipeline):
    def __init__(self, config):
        super(SparkMLPipeline, self).__init__()

        verify_config(config)

        self.config = config

        self.pipe_builder = get_pipeline_builder(config['pipe']['type'])(config)
        self._pipeline, self.label_col = self.pipe_builder.build_pipeline()

    def fit(self, training_data):
        verify_input_dataset(self.config, training_data)
        model = self._pipeline.fit(training_data)

        return model

    def fit_with_outputs(self, training_data):
        verify_input_dataset(self.config, training_data)
        model, y_true, y_pred, feature_importances = \
            fit_pipeline(self._pipeline, self.config, training_data, self.label_col)

        return model, y_true, y_pred, feature_importances

    def get_metrics(self, y_true, y_pred):
        return self.pipe_builder.get_metrics(y_true, y_pred)


class SparkStatPipeline:
    def __init__(self, config, data):
        verify_config(config)
        verify_input_dataset(config, data)

        self.config = config

        pipe_builder = get_pipeline_builder(config['pipe']['type'])
        self._pipeline = pipe_builder(config, data)

    def get_stat(self):
        stat = self._pipeline.compute_stat()

        return stat



# -*- encoding: utf-8 -*-
from abc import ABCMeta

from pyspark.ml import Pipeline
from pyspark.ml.feature import MinMaxScaler, StandardScaler, MaxAbsScaler, Imputer, VectorAssembler
from pyspark.ml.feature import StringIndexer, OneHotEncoderEstimator
from pyspark.ml.feature import PCA, ChiSqSelector, VectorIndexer

from sparkmlpipe.utils.exceptions import InvalidConfigError, MissingConfigError
from sparkmlpipe.utils.constants import STRING_TYPE, NUM_TYPES, \
    MIN_MAX_SCALER_TYPE, STANDARD_SCALER_TYPE, MAXABS_SCALER_TYPE, \
    MEAN_IMPUTER_TYPE, MEDIAN_IMPUTER_TYPE, CHISQ_FEAT_TYPE, PCA_FEAT_TYPE, DEFAULT_K


class BasePipelineBuilder:
    __metaclass__ = ABCMeta

    def __init__(self, config):
        self.config = config

        self.num_out_cols = []
        self.str_out_cols = []
        self.feat_col = None
        self.label_col = None

    def build_pipeline(self):
        data_prep_stages = self._get_data_prep_stages()
        feat_prep_stages = self._get_feat_prep_stages()
        model_stages = self._get_model_stages()

        pipeline = Pipeline(stages=data_prep_stages + feat_prep_stages + model_stages)
        return pipeline, self.label_col

    def get_metrics(self, y_true, y_pred):
        raise NotImplementedError()

    def _get_data_prep_stages(self):
        header = self.config['data']['header']
        types = self.config['data']['types']

        # preprocess label
        label_stages = []
        if 'label' not in self.config['pipe']:
            raise MissingConfigError('pipe.label')
        label_index = self.config['pipe']['label']
        label_input_col_name = header[label_index]
        label_type = types[label_index]

        if label_type == STRING_TYPE:
            self.label_col = 'indexed_' + label_input_col_name
            label_string_indexer = StringIndexer(inputCol=label_input_col_name,
                                                 outputCol=self.label_col)
            label_stages.append(label_string_indexer)
        elif label_type in NUM_TYPES:
            self.label_col = label_input_col_name
        else:
            raise InvalidConfigError('pipe.label', label_type)

        # preprocess numeric cols
        num_stages = []
        num_cols = [header[i] for i, typ in enumerate(types) if typ in NUM_TYPES and i != label_index]
        self.num_out_cols = num_cols

        imputer_conf = self.config['pipe']['data_prep']['numeric']['imputer']
        if len(num_cols) != 0 and imputer_conf != 0:
            imputer_cols = ['imputed_' + col for col in num_cols]
            if imputer_conf == MEAN_IMPUTER_TYPE:
                num_imputer = Imputer(inputCols=num_cols,
                                       outputCols=imputer_cols).setStrategy('mean')
            elif imputer_conf == MEDIAN_IMPUTER_TYPE:
                num_imputer = Imputer(inputCols=num_cols,
                                       outputCols=imputer_cols).setStrategy('median')
            else:
                raise InvalidConfigError('pipe.data_prep.numeric.imputer', imputer_conf)
            num_stages.append(num_imputer)
            self.num_out_cols = imputer_cols

        scaler_conf = self.config['pipe']['data_prep']['numeric']['scaler']
        if len(num_cols) != 0 and scaler_conf != 0:
            assembled_num_col_name = 'assembled_num_features'
            scaled_features_col_name = 'scaled_features'
            scaler_assembler = VectorAssembler(inputCols=self.num_out_cols, outputCol=assembled_num_col_name)
            if scaler_conf == MIN_MAX_SCALER_TYPE:
                num_scaler = MinMaxScaler(inputCol=assembled_num_col_name, outputCol=scaled_features_col_name)
            elif scaler_conf == STANDARD_SCALER_TYPE:
                num_scaler = StandardScaler(inputCol=assembled_num_col_name, outputCol=scaled_features_col_name)
            elif scaler_conf == MAXABS_SCALER_TYPE:
                num_scaler = MaxAbsScaler(inputCol=assembled_num_col_name, outputCol=scaled_features_col_name)
            else:
                raise InvalidConfigError('pipe.data_prep.numeric.scaler', scaler_conf)

            num_stages.append(scaler_assembler)
            num_stages.append(num_scaler)
            # overwrite the num out cols
            self.num_out_cols = [scaled_features_col_name]

        # preprocess string cols
        str_stages = []
        str_cols = [header[i] for i, typ in enumerate(types) if typ == STRING_TYPE and i != label_index]
        str_indexed_cols = [str_col + '_indexed' for str_col in str_cols]
        onehot_conf = self.config['pipe']['data_prep']['string']['onehot']
        if len(str_cols) != 0:
            if onehot_conf == 0:
                # str_indexer
                str_indexers = [
                    StringIndexer(inputCol=str_col,
                                  outputCol=str_out_col)
                    for str_col, str_out_col in zip(str_cols, str_indexed_cols)
                ]

                str_stages.extend(str_indexers)
                self.str_out_cols = str_indexed_cols

            elif onehot_conf == 1:
                # str_indexer
                str_indexers = [
                    StringIndexer(inputCol=str_col,
                                  outputCol=str_out_col)
                    for str_col, str_out_col in zip(str_cols, str_indexed_cols)
                ]

                # onehot_encoder
                str_vec_cols = [str_indexed_col + '_vec' for str_indexed_col in str_indexed_cols]
                str_encoder = OneHotEncoderEstimator(inputCols=str_indexed_cols,
                                                 outputCols=str_vec_cols)

                str_stages.extend(str_indexers)
                str_stages.append(str_encoder)
                self.str_out_cols = str_vec_cols
            else:
                raise InvalidConfigError('pipe.data_prep.string.onehot', onehot_conf)

        return label_stages + num_stages + str_stages

    def _get_feat_prep_stages(self):
        feat_stages = []

        feat_name = self.config['pipe']['feat_prep']['name']

        input_cols = self.num_out_cols + self.str_out_cols

        if input_cols is []:
            raise RuntimeError('the input columns cannot be empty')

        feat_assembler = VectorAssembler(inputCols=input_cols, outputCol=feat_name)

        feat_stages.append(feat_assembler)

        feat_type = self.config['pipe']['feat_prep']['type']

        if feat_type != 0:
            if 'k' in self.config['pipe']['feat_prep']:
                feat_k = self.config['pipe']['feat_prep']['k']
            else:
                feat_k = DEFAULT_K

            if feat_type == CHISQ_FEAT_TYPE:
                selector = ChiSqSelector(numTopFeatures=feat_k, featuresCol=feat_name,
                                         outputCol='selected_' + feat_name, labelCol=self.label_col)

                feat_stages.append(selector)
                self.feat_col = 'selected_' + feat_name
            elif feat_type == PCA_FEAT_TYPE:
                pca = PCA(k=feat_k, inputCol=feat_name, outputCol='pca_' + feat_name)

                feat_stages.append(pca)
                self.feat_col = 'pca_' + feat_name
            else:
                raise InvalidConfigError('pipe.feat_prep.type', feat_type)
        else:
            self.feat_col = feat_name

        feat_index = self.config['pipe']['feat_prep']['index']
        if feat_index != 0:
            indexer = VectorIndexer(inputCol=self.feat_col,
                                    outputCol='indexed_' + self.feat_col, maxCategories=feat_index)
            self.feat_col = 'indexed_' + self.feat_col
            feat_stages.append(indexer)

        return feat_stages

    def _get_model_stages(self):
        raise NotImplementedError()



# -*- encoding: utf-8 -*-
from abc import ABCMeta

from pyspark.ml import Pipeline
from pyspark.ml.feature import MinMaxScaler, StandardScaler, MaxAbsScaler, Imputer, VectorAssembler

from pyspark.ml.feature import StringIndexer, OneHotEncoderEstimator

from sparkmlpipe.utils.exceptions import InvalidConfigError
from sparkmlpipe.utils.constants import STRING_TYPE, NUM_TYPES, \
    MIN_MAX_SCALER_TYPE, STANDARD_SCALER_TYPE, MAXABS_SCALER_TYPE, \
    MEAN_IMPUTER_TYPE, MEDIAN_IMPUTER_TYPE


class BaseStatsBuilder:
    __metaclass__ = ABCMeta

    def __init__(self, config, data):
        self.config = config
        self.data = data
        self.num_out_cols = []
        self.str_out_cols = []
        self.feat_col = None


    def compute_stat(self):
        data_prep_stages = self._get_data_prep_stages()
        feat_prep_stages = self._get_feat_prep_stages()
        pipeline = Pipeline(stages=data_prep_stages + feat_prep_stages)

        prep_pipe = pipeline.fit(self.data)
        transformed_data = prep_pipe.transform(self.data)

        return self._get_stat(transformed_data)

    def _get_data_prep_stages(self):
        header = self.config['data']['header']
        types = self.config['data']['types']

        # preprocess numeric cols
        num_stages = []
        num_cols = [header[i] for i, typ in enumerate(types) if typ in NUM_TYPES]
        self.num_out_cols = num_cols

        imputer_conf = self.config['pipe']['data_prep']['numeric']['imputer']
        if len(num_cols) != 0 and imputer_conf != 0:
            imputer_cols = ['imputed_' + col for col in num_cols]
            if imputer_conf == MEAN_IMPUTER_TYPE:
                num_imputer = Imputer(inputCols=num_cols,
                                      outputCols=imputer_cols).setStrategy('mean')
            elif imputer_conf == MEDIAN_IMPUTER_TYPE:
                num_imputer = Imputer(inputCols=num_cols,
                                      outputCols=imputer_cols).setStrategy('median')
            else:
                raise InvalidConfigError('pipe.data_prep.numeric.imputer', imputer_conf)
            num_stages.append(num_imputer)
            self.num_out_cols = imputer_cols

        scaler_conf = self.config['pipe']['data_prep']['numeric']['scaler']
        if len(num_cols) != 0 and scaler_conf != 0:
            assembled_num_col_name = 'assembled_num_features'
            scaled_features_col_name = 'scaled_features'
            scaler_assembler = VectorAssembler(inputCols=self.num_out_cols, outputCol=assembled_num_col_name)
            if scaler_conf == MIN_MAX_SCALER_TYPE:
                num_scaler = MinMaxScaler(inputCol=assembled_num_col_name, outputCol=scaled_features_col_name)
            elif scaler_conf == STANDARD_SCALER_TYPE:
                num_scaler = StandardScaler(inputCol=assembled_num_col_name, outputCol=scaled_features_col_name)
            elif scaler_conf == MAXABS_SCALER_TYPE:
                num_scaler = MaxAbsScaler(inputCol=assembled_num_col_name, outputCol=scaled_features_col_name)
            else:
                raise InvalidConfigError('pipe.data_prep.numeric.scaler', scaler_conf)

            num_stages.append(scaler_assembler)
            num_stages.append(num_scaler)
            # overwrite the num out cols
            self.num_out_cols = [scaled_features_col_name]

        # preprocess string cols
        str_stages = []
        str_cols = [header[i] for i, typ in enumerate(types) if typ == STRING_TYPE]
        str_indexed_cols = [str_col + '_indexed' for str_col in str_cols]
        onehot_conf = self.config['pipe']['data_prep']['string']['onehot']
        if len(str_cols) != 0:
            if onehot_conf == 0:
                # str_indexer
                str_indexers = [
                    StringIndexer(inputCol=str_col,
                                  outputCol=str_out_col)
                    for str_col, str_out_col in zip(str_cols, str_indexed_cols)
                ]

                str_stages.extend(str_indexers)
                self.str_out_cols = str_indexed_cols

            elif onehot_conf == 1:
                # str_indexer
                str_indexers = [
                    StringIndexer(inputCol=str_col,
                                  outputCol=str_out_col)
                    for str_col, str_out_col in zip(str_cols, str_indexed_cols)
                ]

                # onehot_encoder
                str_vec_cols = [str_indexed_col + '_vec' for str_indexed_col in str_indexed_cols]
                str_encoder = OneHotEncoderEstimator(inputCols=str_indexed_cols,
                                                 outputCols=str_vec_cols)

                str_stages.extend(str_indexers)
                str_stages.append(str_encoder)
                self.str_out_cols = str_vec_cols
            else:
                raise InvalidConfigError('pipe.data_prep.string.onehot', onehot_conf)

        return num_stages + str_stages

    def _get_feat_prep_stages(self):
        feat_stages = []

        feat_name = self.config['pipe']['feat_prep']['name']

        input_cols = self.num_out_cols + self.str_out_cols

        if input_cols is []:
            raise RuntimeError('the input columns cannot be empty')

        feat_assembler = VectorAssembler(inputCols=input_cols, outputCol=feat_name)

        feat_stages.append(feat_assembler)

        self.feat_col = feat_name

        return feat_stages

    def _get_stat(self, data):
        raise NotImplementedError()



# -*- encoding: utf-8 -*-
from sparkmlpipe.utils import dependencies
from sparkmlpipe.__version__ import __version__


__MANDATORY_PACKAGES__ = '''
pyspark>=2.3.0
PyYAML>=3.12
'''

dependencies.verify_packages(__MANDATORY_PACKAGES__)



# -*- encoding: utf-8 -*-
from pyspark.ml.stat import Correlation

from .base import BaseStatsBuilder
from sparkmlpipe.utils.exceptions import InvalidConfigError
from sparkmlpipe.utils.constants import PEARSON_CORR, SPEARMAN_CORR


class CorrelationBuilder(BaseStatsBuilder):

    def _get_stat(self, data):
        model_conf = self.config['pipe']['model']
        params = model_conf['params'] if 'params' in model_conf else {}

        if model_conf['type'] == PEARSON_CORR:
            corr = Correlation.corr(data, self.feat_col, 'pearson').collect()[0][0]
        elif model_conf['type'] == SPEARMAN_CORR:
            corr = Correlation.corr(data, self.feat_col, 'spearman').collect()[0][0]
        else:
            raise InvalidConfigError('pipe.model.type', model_conf['type'])
        return corr



# -*- encoding: utf-8 -*-
import os
from setuptools import setup, find_packages

requirements = [
    "pyspark>=2.3.0",
    "PyYAML>=3.12",
]

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as f:
    long_description = f.read()


with open("sparkmlpipe/__version__.py") as fh:
    version = fh.readlines()[-1].split()[-1].strip("\"'")

setup(
    name='sparkml-pipe',
    description='A configurable PySpark pipeline library.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version=version,
    packages=find_packages(exclude=['test']),
    install_requires=requirements,
    setup_requires=['setuptools>=38.6.0'],
    include_package_data=True,
    author='Michi Tech',
    author_email='parkerzf@gmail.com',
    license='BSD',
    platforms=['Linux'],
    classifiers=[],
    url='https://bitbucket.org/xtechai/sparkml-pipe')



# -*- encoding: utf-8 -*-
class MissingPackageError(Exception):
    error_message = 'mandatory package \'{name}\' not found'

    def __init__(self, package_name):
        self.package_name = package_name
        super(MissingPackageError, self).__init__(
            self.error_message.format(name=package_name))


class IncorrectPackageVersionError(Exception):
    error_message = '\'{name} {installed_version}\' version mismatch ({operation}{required_version})'

    def __init__(self, package_name, installed_version, operation,
                 required_version):
        self.package_name = package_name
        self.installed_version = installed_version
        self.operation = operation
        self.required_version = required_version
        message = self.error_message.format(name=package_name,
                                            installed_version=installed_version,
                                            operation=operation,
                                            required_version=required_version)
        super(IncorrectPackageVersionError, self).__init__(message)


###########################################
###########################################
class MismatchTypeError(Exception):
    error_message = 'Type \'{type}\' and \'{value}\' mismatched'

    def __init__(self, conf_type, conf_value):
        self.conf_type = conf_type
        self.conf_value = conf_value
        super(MismatchTypeError, self).__init__(
            self.error_message.format(type=conf_type, value=conf_value))


class InvalidConfigError(Exception):
    error_message = 'Config \'{name}={value}\' invalid'

    def __init__(self, conf_name, conf_value):
        self.conf_name = conf_name
        self.conf_value = conf_value
        super(InvalidConfigError, self).__init__(
            self.error_message.format(name=conf_name, value=conf_value))


class MismatchColumnNumberError(Exception):
    error_message = 'Config #columns=\'{num_conf_col}\' and Data #columns=\'{num_data_col}\' mismatched'

    def __init__(self, num_data_col, num_conf_col):
        self.num_data_col = num_data_col
        self.num_conf_col = num_conf_col
        message = self.error_message.format(num_conf_col=num_conf_col,
                                            num_data_col=num_data_col,)
        super(MismatchColumnNumberError, self).__init__(message)


class MissingConfigError(Exception):
    error_message = 'Mandatory config \'{name}\' not found'

    def __init__(self, conf_name):
        self.conf_name = conf_name
        super(MissingConfigError, self).__init__(
            self.error_message.format(name=conf_name))